from datetime import date, time, timedelta
from decimal import Decimal
import itertools

from django.utils import timezone
from six.moves import xrange

from .models import Person

def get_fixtures(n=None):
    """
      Returns `n` dictionaries of `Person` objects.
      If `n` is not specified it defaults to 6.
    """
    _now = timezone.now().replace(microsecond=0)  # mysql doesn't do microseconds. # NOQA
    _date = date(2015, 3, 28)
    _time = time(13, 0)
    fixtures = [
        {
            'big_age': 59999999999999999, 'comma_separated_age': '1,2,3',
            'age': -99, 'positive_age': 9999, 'positive_small_age': 299,
            'small_age': -299, 'certified': False, 'null_certified': None,
            'name': 'Mike', 'email': 'miketakeahike@mailinator.com',
            'file_path': '/Users/user/fixtures.json', 'slug': 'mike',
            'text': 'here is a dummy text',
            'url': 'https://docs.djangoproject.com',
            'height': Decimal('1.81'), 'date_time': _now,
            'date': _date, 'time': _time, 'float_height': 0.3,
            'remote_addr': '192.0.2.30', 'my_file': 'dummy.txt',
            'image': 'kitten.jpg', 'data': {'name': 'Mike', 'age': -99},
        },
        {
            'big_age': 245999992349999, 'comma_separated_age': '6,2,9',
            'age': 25, 'positive_age': 49999, 'positive_small_age': 315,
            'small_age': 5409, 'certified': False, 'null_certified': True,
            'name': 'Pete', 'email': 'petekweetookniet@mailinator.com',
            'file_path': 'users.json', 'slug': 'pete', 'text': 'dummy',
            'url': 'https://google.com', 'height': Decimal('1.93'),
            'date_time': _now, 'date': _date, 'time': _time,
            'float_height': 0.5, 'remote_addr': '127.0.0.1',
            'my_file': 'fixtures.json',
            'data': [{'name': 'Pete'}, {'name': 'Mike'}],
        },
        {
            'big_age': 9929992349999, 'comma_separated_age': '6,2,9,10,5',
            'age': 29, 'positive_age': 412399, 'positive_small_age': 23315,
            'small_age': -5409, 'certified': False, 'null_certified': True,
            'name': 'Ash', 'email': 'rashash@mailinator.com',
            'file_path': '/Downloads/kitten.jpg', 'slug': 'ash',
            'text': 'bla bla bla', 'url': 'news.ycombinator.com',
            'height': Decimal('1.78'), 'date_time': _now,
            'date': _date, 'time': _time,
            'float_height': 0.8, 'my_file': 'dummy.png',
            'data': {'text': 'bla bla bla', 'names': ['Mike', 'Pete']},
        },
        {
            'big_age': 9992349234, 'comma_separated_age': '12,29,10,5',
            'age': -29, 'positive_age': 4199, 'positive_small_age': 115,
            'small_age': 909, 'certified': True, 'null_certified': False,
            'name': 'Mary', 'email': 'marykrismas@mailinator.com',
            'file_path': 'dummy.png', 'slug': 'mary',
            'text': 'bla bla bla bla bla', 'url': 'news.ycombinator.com',
            'height': Decimal('1.65'), 'date_time': _now,
            'date': _date, 'time': _time, 'float_height': 0,
            'remote_addr': '2a02:42fe::4',
            'data': {'names': {'name': 'Mary'}},
        },
        {
            'big_age': 999234, 'comma_separated_age': '12,1,30,50',
            'age': 1, 'positive_age': 99199, 'positive_small_age': 5,
            'small_age': -909, 'certified': False, 'null_certified': False,
            'name': 'Sandra', 'email': 'sandrasalamandr@mailinator.com',
            'file_path': '/home/dummy.png', 'slug': 'sandra',
            'text': 'this is a dummy text', 'url': 'google.com',
            'height': Decimal('1.59'), 'date_time': _now,
            'date': _date, 'time': _time, 'float_height': 2 ** 2,
            'image': 'dummy.jpeg', 'data': {},
        },
        {
            'big_age': 9999999999, 'comma_separated_age': '1,100,3,5',
            'age': 35, 'positive_age': 1111, 'positive_small_age': 500,
            'small_age': 110, 'certified': True, 'null_certified': None,
            'name': 'Crystal', 'email': 'crystalpalace@mailinator.com',
            'file_path': '/home/dummy.txt', 'slug': 'crystal',
            'text': 'dummy text', 'url': 'docs.djangoproject.com',
            'height': Decimal('1.71'), 'date_time': _now,
            'date': _date, 'time': _time, 'float_height': 2 ** 10,
            'image': 'dummy.png', 'data': [],
        },
    ]
    n = n or len(fixtures)
    fixtures = itertools.cycle(fixtures)
    for _ in xrange(n):
        yield next(fixtures)


def create_fixtures(n=None):
    """
      Wrapper for Person.bulk_create which creates `n` fixtures
    """
    Person.objects.bulk_create(Person(**person)
                               for person in get_fixtures(n))



from datetime import date, time, timedelta
from decimal import Decimal
import random

from django.test import TestCase
from django.utils import timezone

from .models import Person
from .fixtures import create_fixtures


class BulkUpdateTests(TestCase):

    def setUp(self):
        self.now = timezone.now().replace(microsecond=0)  # mysql doesn't do microseconds. # NOQA
        self.date = date(2015, 3, 28)
        self.time = time(13, 0)
        create_fixtures()

    def _test_field(self, field, idx_to_value_function):
        '''
        Helper to do repeative simple tests on one field.
        '''

        # set
        people = Person.objects.order_by('pk').all()
        for idx, person in enumerate(people):
            value = idx_to_value_function(idx)
            setattr(person, field, value)

        # update
        Person.objects.bulk_update(people, update_fields=[field])

        # check
        people = Person.objects.order_by('pk').all()
        for idx, person in enumerate(people):
            saved_value = getattr(person, field)
            expected_value = idx_to_value_function(idx)
            self.assertEqual(saved_value, expected_value)

    def test_simple_fields(self):
        fn = lambda idx: idx + 27
        for field in ('default', 'big_age', 'age', 'positive_age',
                      'positive_small_age', 'small_age'):
            self._test_field(field,  fn)

    def test_comma_separated_integer_field(self):
        fn = lambda idx: str(idx) + ',27'
        self._test_field('comma_separated_age',  fn)

    def test_boolean_field(self):
        fn = lambda idx: [True, False][idx % 2]
        self._test_field('certified',  fn)

    def test_null_boolean_field(self):
        fn = lambda idx: [True, False, None][idx % 3]
        self._test_field('null_certified',  fn)

    def test_char_field(self):
        NAMES = ['Walter', 'The Dude', 'Donny', 'Jesus', 'Buddha', 'Clark']
        fn = lambda idx: NAMES[idx % 5]
        self._test_field('name',  fn)

    def test_email_field(self):
        EMAILS = ['walter@mailinator.com', 'thedude@mailinator.com',
                  'donny@mailinator.com', 'jesus@mailinator.com',
                  'buddha@mailinator.com', 'clark@mailinator.com']
        fn = lambda idx: EMAILS[idx % 5]
        self._test_field('email',  fn)

    def test_file_path_field(self):
        PATHS = ['/home/dummy.txt', '/Downloads/kitten.jpg',
                 '/Users/user/fixtures.json', 'dummy.png',
                 'users.json', '/home/dummy.png']
        fn = lambda idx: PATHS[idx % 5]
        self._test_field('file_path',  fn)

    def test_slug_field(self):
        SLUGS = ['jesus', 'buddha', 'clark', 'the-dude', 'donny', 'walter']
        fn = lambda idx: SLUGS[idx % 5]
        self._test_field('slug',  fn)

    def test_text_field(self):
        TEXTS = ['this is a dummy text', 'dummy text', 'bla bla bla bla bla',
                 'here is a dummy text', 'dummy', 'bla bla bla']
        fn = lambda idx: TEXTS[idx % 5]
        self._test_field('text',  fn)

    def test_url_field(self):
        URLS = ['docs.djangoproject.com', 'news.ycombinator.com',
                'https://docs.djangoproject.com', 'https://google.com',
                'google.com', 'news.ycombinator.com']
        fn = lambda idx: URLS[idx % 5]
        self._test_field('url',  fn)

    def test_date_time_field(self):
        fn = lambda idx: self.now - timedelta(days=1 + idx, hours=1 + idx)
        self._test_field('date_time',  fn)

    def test_date_field(self):
        fn = lambda idx: self.date - timedelta(days=1 + idx)
        self._test_field('date',  fn)

    def test_time_field(self):
        fn = lambda idx: time(1 + idx, idx)
        self._test_field('time',  fn)

    def test_decimal_field(self):
        fn = lambda idx: Decimal('1.%s' % (50 + idx * 7))
        self._test_field('height',  fn)

    def test_float_field(self):
        fn = lambda idx: float(idx) * 2.0
        self._test_field('float_height',  fn)

    def test_generic_ipaddress_field(self):
        IPS = ['127.0.0.1', '192.0.2.30', '2a02:42fe::4', '10.0.0.1',
               '8.8.8.8']
        fn = lambda idx: IPS[idx % 5]
        self._test_field('remote_addr',  fn)

    def test_image_field(self):
        IMGS = ['kitten.jpg', 'dummy.png', 'user.json', 'dummy.png', 'foo.gif']
        fn = lambda idx: IMGS[idx % 5]

        self._test_field('image',  fn)
        self._test_field('my_file',  fn)

    def test_custom_fields(self):
        values = {}
        people = Person.objects.all()
        people_dict = {p.name: p for p in people}
        person = people_dict['Mike']
        person.data = {'name': 'mikey', 'age': 99, 'ex': -99}
        values[person.pk] = {'name': 'mikey', 'age': 99, 'ex': -99}

        person = people_dict['Mary']
        person.data = {'names': {'name': []}}
        values[person.pk] = {'names': {'name': []}}

        person = people_dict['Pete']
        person.data = []
        values[person.pk] = []

        person = people_dict['Sandra']
        person.data = [{'name': 'Pete'}, {'name': 'Mike'}]
        values[person.pk] = [{'name': 'Pete'}, {'name': 'Mike'}]

        person = people_dict['Ash']
        person.data = {'text': 'bla'}
        values[person.pk] = {'text': 'bla'}

        person = people_dict['Crystal']
        values[person.pk] = person.data

        Person.objects.bulk_update(people)

        people = Person.objects.all()
        for person in people:
            self.assertEqual(person.data, values[person.pk])

    def test_update_fields(self):
        """
            Only the fields in "update_fields" are updated
        """
        people = Person.objects.order_by('pk').all()
        for idx, person in enumerate(people):
            person.age += 1
            person.height += Decimal('0.01')
        Person.objects.bulk_update(people, update_fields=['age'])

        people2 = Person.objects.order_by('pk').all()
        for person1, person2 in zip(people, people2):
            self.assertEqual(person1.age, person2.age)
            self.assertNotEqual(person1.height, person2.height)

    def test_exclude_fields(self):
        """
            Only the fields not in "exclude_fields" are updated
        """
        people = Person.objects.order_by('pk').all()
        for idx, person in enumerate(people):
            person.age += 1
            person.height += Decimal('0.01')
        Person.objects.bulk_update(people, exclude_fields=['age'])

        people2 = Person.objects.order_by('pk').all()
        for person1, person2 in zip(people, people2):
            self.assertNotEqual(person1.age, person2.age)
            self.assertEqual(person1.height, person2.height)

    def test_object_list(self):
        """
          Pass in a list instead of a queryset for bulk updating
        """
        people = Person.objects.order_by('pk').all()
        for idx, person in enumerate(people):
            person.big_age = idx + 27
        Person.objects.bulk_update(list(people))

        people = Person.objects.order_by('pk').all()
        for idx, person in enumerate(people):
            self.assertEqual(person.big_age, idx + 27)

    def test_empty_list(self):
        """
        Update no elements, passed as a list
        """
        Person.objects.bulk_update([])

    def test_empty_queryset(self):
        """
        Update no elements, passed as a queryset
        """
        Person.objects.bulk_update(Person.objects.filter(name="Aceldotanrilsteucsebces ECSbd (funny name, isn't it?)"))




import os
import dj_database_url

BASE_DIR = os.path.join(os.path.dirname(os.path.dirname(__file__)), os.pardir)

SECRET_KEY = 'not_so_secret_r2fetCebUAZB3DhzTn5NPg8J08IancuGt04npTMh'
DEBUG = True

db_url = os.environ.get("DATABASE_URL", "sqlite://localhost/:memory:")
DB = dj_database_url.parse(db_url)

DATABASES = {
    'default': DB,
}

INSTALLED_APPS = ('bulk_update', 'tests',)
MIDDLEWARE_CLASSES = ()



# coding: utf8
"""
Main module with the bulk_update function.
"""

import itertools

from django.db import connections, models
from django.db.models.query import QuerySet


def _get_db_type(field, connection):
    if isinstance(field, (models.PositiveSmallIntegerField,
                          models.PositiveIntegerField)):
        # integer CHECK ("points" >= 0)'
        return field.db_type(connection).split(' ', 1)[0]

    res = field.db_type(connection)
    return res


def grouper(iterable, size):
    # http://stackoverflow.com/a/8991553
    it = iter(iterable)
    if size <= 0:
        yield it
        return
    while True:
        chunk = tuple(itertools.islice(it, size))
        if not chunk:
            return
        yield chunk


def bulk_update(objs, meta=None, update_fields=None, exclude_fields=None,
                using='default', batch_size=None, pk_field='pk'):
    assert batch_size is None or batch_size > 0

    # if we have a QuerySet, avoid loading objects into memory
    if isinstance(objs, QuerySet):
        if not objs.exists():
            return
        batch_size = batch_size or objs.count()
    else:
        if not objs:
            return
        batch_size = batch_size or len(objs)

    connection = connections[using]
    if meta is None:
        # TODO: account for iterables
        meta = objs[0]._meta

    if pk_field == 'pk':
        pk_field = meta.pk.name

    exclude_fields = exclude_fields or []
    update_fields = update_fields or meta.get_all_field_names()
    fields = [
        f for f in meta.fields
        if ((not isinstance(f, models.AutoField))
            and (f.attname in update_fields))]
    fields = [
        f for f in fields
        if f.attname not in exclude_fields]

    # The case clause template; db-dependent
    # Apparently, mysql's castable types are very limited and have
    # nothing to do with the column types. Still, it handles the uncast
    # types well enough... hopefully.
    # http://dev.mysql.com/doc/refman/5.5/en/cast-functions.html#function_cast
    #
    # Sqlite also gives some trouble with cast, at least for datetime,
    # but is also permissive for uncast values
    vendor = connection.vendor
    use_cast = 'mysql' not in vendor and 'sqlite' not in connection.vendor
    if use_cast:
        case_clause_template = '"{column}" = CAST(CASE "{pkcolumn}" {{when}}'
        tail_end_template = 'END AS {type})'
    else:
        case_clause_template = '"{column}" = (CASE "{pkcolumn}" {{when}}'
        tail_end_template = 'END)'

    # String escaping in ANSI SQL is done by using double quotes (").
    # Unfortunately, this escaping method is not portable to MySQL, unless it
    # is set in ANSI compatibility mode.
    quote_mark = '"' if 'mysql' not in vendor else '`'
    case_clause_template = case_clause_template.replace('"', quote_mark)

    for objs_batch in grouper(objs, batch_size):
        pks = []
        case_clauses = {}
        for obj in objs_batch:
            pks.append(getattr(obj, pk_field))
            for field in fields:
                column = field.column

                # Synopsis: make sure the column-specific 'case'
                # exists and then append the obj-specific values to
                # it in a tricky way (leaving ' {when}' at the end).
                # TODO?: For speed, use a list to be ''-joined later,
                # instead? Or a bytearray.
                # TODO: optimise (getitem+setitem vs. get + setitem)
                try:
                    case_clause = case_clauses[column]
                except KeyError:
                    case_clause = {
                        'sql': case_clause_template.format(
                            column=column, pkcolumn=meta.get_field(pk_field).column),
                        'params': [],
                        'type': _get_db_type(field, connection=connection),
                    }
                    case_clauses[column] = case_clause

                case_clause['sql'] = (
                    case_clause['sql'].format(when="WHEN %s THEN %s {when}")
                )

                case_clause['params'].extend(
                    [getattr(obj, pk_field),
                     field.get_db_prep_value(
                         getattr(obj, field.attname), connection)])

        if pks:
            values = ', '.join(
                v['sql'].format(when=tail_end_template.format(type=v['type']))
                for v in case_clauses.values())
            parameters = [
                param
                for clause in case_clauses.values()
                for param in clause['params']]

            del case_clauses  # ... memory

            pkcolumn = meta.get_field(pk_field).column
            dbtable = '{0}{1}{0}'.format(quote_mark, meta.db_table)
            # Storytime: apparently (at least for mysql and sqlite), if a
            # non-simple parameter is added (e.g. a tuple), it is
            # processed with force_text and, accidentally, manages to
            # be a valid syntax... unless there's only one element.
            # So, to fix this, expand the ' in %s' with the parameters' string
            in_clause_sql = '({})'.format(
                ', '.join(itertools.repeat('%s', len(pks))))
            parameters.extend(pks)

            sql = (
                'UPDATE {dbtable} SET {values} WHERE {pkcolumn} '
                'in {in_clause_sql}'
                .format(
                    dbtable=dbtable, values=values, pkcolumn=pkcolumn,
                    in_clause_sql=in_clause_sql))
            del values, pks

            connection.cursor().execute(sql, parameters)



from django.db import models
from .helper import bulk_update


class BulkUpdateManager(models.Manager):
    def bulk_update(self, objs, update_fields=None, exclude_fields=None):
        bulk_update(objs, update_fields=update_fields,
                    exclude_fields=exclude_fields, using=self.db)



def queryset_chunker(queryset, chunk_size=5000, order_by='-pk', limit=None):
    """
    Takes lazy queryset and chunks it by given chunk_size.

    :param queryset: type of queryset. This is a lazy queryset to
        split into chunks
    :type queryset: QuerySet
    :param chunk_size: size of smaller pieces
    :type chunk_size: int
    """
    if queryset.query.distinct:
        order_by = queryset.query.distinct_fields[0]
    if not queryset.query.order_by:
        queryset = queryset.order_by(order_by)
    for pivot in xrange(0, queryset.count(), chunk_size):
        if limit and pivot > limit:
            break
        yield queryset[pivot:pivot + chunk_size]



import os
import sys
from setuptools import setup, find_packages

README = open(os.path.join(os.path.dirname(__file__), 'README.md')).read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

version = "1.1.6"

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    os.system('python setup.py bdist_wheel upload')
    sys.exit()

if sys.argv[-1] == 'tag':
    os.system("git tag -a %s -m 'version %s'" % (version, version))
    os.system("git push --tags")
    sys.exit()

setup(
    name='django-bulk-update',
    version=version,
    packages=find_packages(),
    include_package_data=True,
    description='Bulk update using one query over Django ORM.',
    long_description=README,
    url='https://github.com/aykut/django-bulk-update',
    author='Aykut Ozat',
    author_email='aykutozat@gmail.com',
    zip_safe=False,
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



from django.db import models

from jsonfield import JSONField

from bulk_update.manager import BulkUpdateManager


class Person(models.Model):
    """
        test model
    """
    big_age = models.BigIntegerField()
    comma_separated_age = models.CommaSeparatedIntegerField(max_length=255)
    age = models.IntegerField()
    positive_age = models.PositiveIntegerField()
    positive_small_age = models.PositiveSmallIntegerField()
    small_age = models.SmallIntegerField()
    height = models.DecimalField(max_digits=3, decimal_places=2)
    float_height = models.FloatField()

    certified = models.BooleanField(default=False)
    null_certified = models.NullBooleanField()

    name = models.CharField(max_length=140, blank=True, null=True)
    email = models.EmailField()
    file_path = models.FilePathField()
    slug = models.SlugField()
    text = models.TextField()
    url = models.URLField()

    date_time = models.DateTimeField(null=True, blank=True)
    date = models.DateField(null=True, blank=True)
    time = models.TimeField(null=True, blank=True)

    remote_addr = models.GenericIPAddressField(null=True, blank=True)

    my_file = models.FileField(upload_to='/some/path/', null=True, blank=True)
    image = models.ImageField(upload_to='/some/path/', null=True, blank=True)

    data = JSONField(null=True, blank=True)

    default = models.IntegerField(null=True, blank=True,
                                  help_text="A reserved keyword")

    objects = BulkUpdateManager()
